from itertools import zip_longest, tee

from numpy import all as np_all, arange, isin, sort, concatenate

from hypothesis import given, assume, example
from hypothesis.strategies import one_of

from pytest import raises

from ..chunking import ChunkSize
from ..slice import Slice
from ..tuple import Tuple
from ..ndindex import ndindex

from .helpers import assert_equal, chunk_sizes, chunk_shapes, prod, ints, slices, ndindices

def test_ChunkSize_constructor():
    raises(TypeError, lambda: ChunkSize(Tuple(1, 2, 3)))
    raises(TypeError, lambda: ChunkSize(1, 2, 3))
    raises(TypeError, lambda: ChunkSize(1))
    raises(ValueError, lambda: ChunkSize((-1, 2, 3)))
    raises(ValueError, lambda: ChunkSize((0, 2, 3)))
    raises(NotImplementedError, lambda: ChunkSize((None, 2, 3)))

@given(chunk_sizes())
def test_ChunkSize_eq(chunk_size_tuple):
    chunk_size = ChunkSize(chunk_size_tuple)
    new = type(chunk_size)(*chunk_size.args)

    assert chunk_size == chunk_size_tuple
    assert chunk_size_tuple == chunk_size
    assert new == chunk_size
    assert chunk_size == new
    assert new == chunk_size_tuple
    assert chunk_size_tuple == new

    assert hash(new) == hash(chunk_size)
    assert not (chunk_size == 'a')
    assert not ('a' == chunk_size)
    assert (chunk_size != 'a')
    assert ('a' != chunk_size)

    h = hash(chunk_size_tuple)
    assert hash(chunk_size) == h

@given(chunk_sizes(), one_of(ints(), slices()))
def test_ChunkSize_args(chunk_size_tuple, idx):
    chunk_size = ChunkSize(chunk_size_tuple)
    assert chunk_size.args == (chunk_size_tuple,)

    try:
        ndindex(idx)
    except ValueError: # pragma: no cover
        # Filter out invalid slices (TODO: do this in the strategy)
        assume(False)

    # Should index the same way
    # TODO: Refactor check_same() so we can use that
    try:
        chunk_size_idx = chunk_size[idx]
    except IndexError:
        try:
            tuple_idx = chunk_size_tuple[idx]
        except IndexError:
            pass
        else:
            raise AssertionError("ChunkSize raised but tuple did not")
    else:
        tuple_idx = chunk_size_tuple[idx]
        assert chunk_size_idx == tuple_idx

@given(chunk_sizes())
def test_ChunkSize_tuple(chunk_size_tuple):
    # Test that ChunkSize behaves like a tuple
    chunk_size = ChunkSize(chunk_size_tuple)
    assert tuple(chunk_size) == chunk_size_tuple

def test_indices_error():
    raises(ValueError, lambda: next(ChunkSize((1, 2)).indices((1, 2, 3))))

@given(chunk_sizes(), chunk_shapes)
def test_num_chunks(chunk_size, shape):
    chunk_size = ChunkSize(chunk_size)
    assert chunk_size.num_chunks(shape) == len(list(chunk_size.indices(shape)))

@given(chunk_sizes(), chunk_shapes)
def test_indices(chunk_size, shape):
    chunk_size = ChunkSize(chunk_size)
    indices = chunk_size.indices(shape)
    size = prod(shape)
    a = arange(size).reshape(shape)

    subarrays = []
    for idx in indices:
        # The indices should be fully expanded
        assert idx.expand(shape) == idx
        # Except for indices at the edges, they should index a full chunk
        if not any(s.stop == i for s, i in zip(idx.args, shape)):
            assert idx.newshape(shape) == chunk_size
        # Make sure they can be indexed
        subarrays.append(a[idx.raw])
    # Check that indices together index every element of the array exactly
    # once.
    elements = [i for x in subarrays for i in x.flatten()]
    assert sorted(elements) == list(range(size))

@example(chunk_size=(1, 1), idx=[[False, True], [True, True]],
         shape=(2, 2))
@example(chunk_size=(1,), idx=slice(None, None, -1), shape=(2,))
@example((1,), True, (1,))
@example(chunk_size=(1, 1), idx=slice(1, None, 2), shape=(4, 1))
@example((1,), ..., (0,))
@example((2, 2), (0, 3), (5, 5))
@example((2, 2), (slice(0, 5, 2), slice(0, 5, 3)), (5, 5))
@example((2, 2), ([0, 0],), (5, 5))
@given(chunk_sizes(), ndindices, chunk_shapes)
def test_as_subchunks(chunk_size, idx, shape):
    chunk_size = ChunkSize(chunk_size)
    size = prod(shape)
    a = arange(size).reshape(shape)
    idx = ndindex(idx)

    try:
        idx.reduce(shape)
    except IndexError:
        assume(False)

    full_idx = a[idx.raw]

    subarrays = []
    fast = chunk_size.as_subchunks(idx, shape)
    slow = chunk_size.as_subchunks(idx, shape, _force_slow=True)
    slow2 = chunk_size.as_subchunks(idx, shape, _force_slow=True)
    no_fallback = chunk_size.as_subchunks(idx, shape, _force_slow=False)
    slow_raised_notimplementederror = False
    try:
        next(slow2)
    except StopIteration:
        pass
    except NotImplementedError:
        # The fallback isn't implemented, but the fast case may still be.
        slow, fast = tee(fast, 2)
        slow_raised_notimplementederror = True
    if not slow_raised_notimplementederror:
        # If it works (no NotImplementedError), it shouldn't use the fallback.
        try:
            next(no_fallback)
        except StopIteration:
            pass
    try:
        for c, cslow in zip_longest(fast, slow):
            assert c == cslow
            index = idx.expand(shape).as_subindex(c)
            chunk = a[c.raw]
            subchunk = chunk[index.raw]
            # Not empty
            assert subchunk.size > 0
            # Indexes the right elements (c.f. test_as_subindex)
            assert_equal(subchunk.flatten(), full_idx[isin(full_idx, chunk)])
            subarrays.append(subchunk)
    except NotImplementedError:
        # NotImplementedError should only be allowed from the fallback algorithm
        if not slow_raised_notimplementederror:
            raise
        return

    # Picks all elements
    if subarrays:
        elements = concatenate([x.flatten() for x in subarrays])
    else:
        elements = arange(0)
    assert_equal(sort(elements), sort(full_idx.flatten()))

def test_as_subchunks_error():
    raises(ValueError, lambda: next(ChunkSize((1, 2)).as_subchunks(..., (1, 2, 3))))

@example(chunk_size=(1,), idx=(slice(None), slice(None)), shape=(0,))
@example(chunk_size=(), idx=(), shape=())
@example(chunk_size=(1, 1), idx=[[False, True], [True, True]],
         shape=(2, 2))
@example(chunk_size=(1,), idx=None, shape=(1,))
@example((1,), True, (1,))
@example(chunk_size=(1, 1), idx=slice(1, None, 2), shape=(4, 1))
@example((1,), ..., (0,))
@example((2, 2), (0, 3), (5, 5))
@example((2, 2), (slice(0, 5, 2), slice(0, 5, 3)), (5, 5))
@example((2, 2), ([0, 0],), (5, 5))
@given(chunk_sizes(), ndindices, chunk_shapes)
def test_num_subchunks(chunk_size, idx, shape):
    chunk_size = ChunkSize(chunk_size)
    idx = ndindex(idx)

    try:
        idx.reduce(shape)
    except IndexError:
        assume(False)

    subchunks = chunk_size.as_subchunks(idx, shape)
    try:
        actual_num_subchunks = len(list(subchunks))
        computed_num_subchunks = chunk_size.num_subchunks(idx, shape)
    except NotImplementedError:
        return
    assert computed_num_subchunks == actual_num_subchunks

def test_num_subchunks_error():
    raises(ValueError, lambda: next(ChunkSize((1, 2)).num_subchunks(..., (1, 2, 3))))


@example((2, 2), (0, False), (5, 5))
@example((5,), [0, 7], (15,))
@example((5,), [], (15,))
@given(chunk_sizes(), ndindices, chunk_shapes)
def test_containing_block(chunk_size, idx, shape):
    chunk_size = ChunkSize(chunk_size)
    idx = ndindex(idx)

    size = prod(shape)
    a = arange(size).reshape(shape)

    try:
        idx.reduce(shape)
    except IndexError:
        assume(False)

    try:
        block = chunk_size.containing_block(idx, shape)
    except NotImplementedError:
        return

    assert isinstance(block, Tuple), block
    assert len(block.args) == len(chunk_size)
    assert all(isinstance(s, Slice) for s in block.args)
    assert all(s.start >= 0 and s.start % n == 0 for s, n in zip(block.args, chunk_size)), block
    assert all(s.stop >= 0 and (s.stop == i or s.stop % n == 0) and s.stop <=
               i for s, i, n in zip(block.args, shape, chunk_size)), block
    assert all(s.step == 1 for s in block.args), block

    a_idx = a[idx.raw]
    a_block = a[block.raw]

    assert np_all(isin(a_idx, a_block))

    # Verify that the block is indeed the smallest possible by shrinking it
    # and making sure that misses some of the index. This check doesn't work
    # for empty indices, so those are handled separately. Also shape == ()
    # cannot give an empty tuple index because we want only tuples of slices..
    if (0 in shape or idx.reduce(shape).isempty()) and shape != ():
        assert block.isempty()
    else:
        for i in range(len(block.args)):
            s = block.args[i]
            new_s1 = Slice(s.start + chunk_size[i], s.stop)
            if s.stop == shape[i] and shape[i] % chunk_size[i]:
                new_s2 = Slice(s.start, s.stop - s.stop % chunk_size[i])
            else:
                new_s2 = Slice(s.start, s.stop - chunk_size[i])
            new_block1 = Tuple(*(block.args[:i] + (new_s1,) + block.args[i+1:]))
            new_block2 = Tuple(*(block.args[:i] + (new_s2,) + block.args[i+1:]))
            a_block1 = a[new_block1.raw]
            a_block2 = a[new_block2.raw]
            assert not np_all(isin(a_idx, a_block1))
            assert not np_all(isin(a_idx, a_block2))

def test_containing_block_error():
    raises(ValueError, lambda: ChunkSize((1, 2)).containing_block(..., (1, 2, 3)))



from numpy import arange

from hypothesis import given, example
from hypothesis.strategies import integers, one_of

from ..ndindex import ndindex
from ..tuple import Tuple
from ..integer import Integer
from .helpers import ndindices, check_same, short_shapes, prod

@example(([False], None), (1,))
@example((False, slice(0, 10)), (5, 2))
@example((None, True, 0), (5, 2))
@example((slice(0, 10), [0, -1]), (5, 2))
@example(slice(0, 10), 5)
@example([0, 1, 2], 3)
@example(([0, 1, 2],), 3)
@given(ndindices, one_of(short_shapes, integers(0, 10)))
def test_selected_indices_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    ndindex(idx)

    def raw_func(a, idx):
        return list(a[idx].flat)

    def ndindex_func(a, index):
        indices = list(index.selected_indices(shape))
        for i in indices:
            if len(a.shape) == 1:
                assert isinstance(i, Integer)
            else:
                assert isinstance(i, Tuple)
                assert all(isinstance(j, Integer) for j in i.args)

        return [a[i.raw] for i in indices]

    def assert_equal(a, b):
        assert a == b

    check_same(a, idx, raw_func=raw_func, ndindex_func=ndindex_func,
               assert_equal=assert_equal, same_exception=False)



from hypothesis import given, example
from hypothesis.strategies import one_of, integers

from numpy import arange

from .helpers import ndindices, shapes, MAX_ARRAY_SIZE, check_same, prod

@example([0], (1,))
@example(..., (1, 2, 3))
@example(slice(0, 1), ())
@example(slice(0, 1), (1,))
@example((0, 1), (2, 2))
@example((0,), ())
@example([[1]], (0, 0, 1))
@example(None, ())
@given(ndindices, one_of(shapes, integers(0, MAX_ARRAY_SIZE)))
def test_isvalid_hypothesis(idx, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    def raw_func(a, idx):
        try:
            a[idx]
            return True
        except Warning as w:
            # check_same unconditionally turns this warning into raise
            # IndexError, so we have to handle it separately here.
            if "Out of bound index found. This was previously ignored when the indexing result contained no elements. In the future the index error will be raised. This error occurs either due to an empty slice, or if an array has zero elements even before indexing." in w.args[0]:
                return False
            raise # pragma: no cover
        except IndexError:
            return False

    def ndindex_func(a, index):
        return index.isvalid(a.shape)

    def assert_equal(x, y):
        assert x == y

    check_same(a, idx, raw_func=raw_func, ndindex_func=ndindex_func,
               assert_equal=assert_equal)



"""
Test that ndindex does not depend on NumPy, as long as no functionality
involving array indices is used.

Also test that it does not depend on SymPy (which used to be a hard
dependency, see ndindex/_crt.py).

Because the tests themselves require NumPy, we must test this in a subprocess
where we know that NumPy is not yet imported.

"""

import sys
import subprocess

import pytest

from ndindex import (ndindex, iter_indices, Integer, Slice, Tuple, ellipsis,
                     Newaxis, ChunkSize, BroadcastError, AxisError)

def _test_dependency_ndindex(mod):
    assert mod not in sys.modules

    # Run various functions that shouldn't depend on NumPy and check that they
    # don't import it.

    ndindex(1)
    assert mod not in sys.modules

    ndindex[0:1]
    assert mod not in sys.modules

    ndindex(ndindex(1))
    assert mod not in sys.modules

    ndindex(None)
    assert mod not in sys.modules

    ndindex[1, 0:1]
    assert mod not in sys.modules

    ndindex(...)
    assert mod not in sys.modules

    try:
        ndindex('')
    except IndexError:
        pass
    assert mod not in sys.modules

def _test_dependency_iter_indices(mod):
    assert mod not in sys.modules

    list(iter_indices((1, 2), (2, 1)))
    assert mod not in sys.modules

    list(iter_indices((1, 2), (2, 1), skip_axes=(1,)))
    assert mod not in sys.modules

    try:
        list(iter_indices((2, 3), (3, 2)))
    except BroadcastError:
        pass
    else:
        assert False
    assert mod not in sys.modules

    try:
        list(iter_indices((2, 3), skip_axes=(2,)))
    except AxisError:
        pass
    else:
        assert False
    assert mod not in sys.modules

def _test_dependency_indices(mod):
    for index_type, args in [
            (Integer, (1,)),
            (Slice, (2, 10, 3)),
            (ellipsis, ()),
            (Newaxis, ()),
            (Tuple, (slice(0, 3), 1)),
    ]:
        assert mod not in sys.modules, index_type

        idx = index_type(*args)
        assert mod not in sys.modules, index_type

        idx.raw
        assert mod not in sys.modules, index_type

        if hasattr(idx, '__len__'):
            len(idx)
        assert mod not in sys.modules, index_type

        idx.reduce()
        assert mod not in sys.modules, index_type

        idx.reduce((2, 3))
        assert mod not in sys.modules, index_type

        idx.newshape((2, 3))
        assert mod not in sys.modules, index_type

        try:
            idx.as_subindex(slice(1, 20, 5))
        except NotImplementedError:
            pass
        assert mod not in sys.modules, index_type

        idx.isempty((2, 3))
        assert mod not in sys.modules, index_type

        idx == idx
        assert mod not in sys.modules, index_type

        idx == 1
        assert mod not in sys.modules, index_type

        hash(idx)
        assert mod not in sys.modules, index_type

        idx.broadcast_arrays()
        assert mod not in sys.modules, index_type

        idx.expand((2, 3))
        assert mod not in sys.modules, index_type

        str(idx)
        assert mod not in sys.modules, index_type

        repr(idx)
        assert mod not in sys.modules, index_type

def _test_dependency_ChunkSize(mod):
    assert mod not in sys.modules

    c = ChunkSize((10, 10))
    assert mod not in sys.modules

    hash(c)
    assert mod not in sys.modules

    c[0]
    assert mod not in sys.modules

    len(c)
    assert mod not in sys.modules

    c.num_chunks((100, 100))
    assert mod not in sys.modules

    list(c.indices((100, 100)))
    assert mod not in sys.modules

    list(c.as_subchunks(slice(0, 50), (100, 100)))
    assert mod not in sys.modules

    c.num_subchunks(slice(0, 50), (100, 100))
    assert mod not in sys.modules

    c.containing_block(slice(0, 50), (100, 100))
    assert mod not in sys.modules

    list(c.as_subchunks(None, (100, 100)))
    assert mod not in sys.modules

    c.num_subchunks(None, (100, 100))
    assert mod not in sys.modules

    c.containing_block(None, (100, 100))
    assert mod not in sys.modules

    list(c.as_subchunks(1, (100, 100)))
    assert mod not in sys.modules

    c.num_subchunks(1, (100, 100))
    assert mod not in sys.modules

    c.containing_block(1, (100, 100))
    assert mod not in sys.modules

    list(c.as_subchunks(..., (100, 100)))
    assert mod not in sys.modules

    c.num_subchunks(..., (100, 100))
    assert mod not in sys.modules

    c.containing_block(..., (100, 100))
    assert mod not in sys.modules

    list(c.as_subchunks((0, slice(0, 50)), (100, 100)))
    assert mod not in sys.modules

    c.num_subchunks((0, slice(0, 50)), (100, 100))
    assert mod not in sys.modules

    c.containing_block((0, slice(0, 50)), (100, 100))
    assert mod not in sys.modules

@pytest.mark.no_cover
@pytest.mark.parametrize('func', [func for func in globals() if func.startswith('_test')])
def test_numpy_dependency(func):
    # This unfortunately won't go through any of the pytest machinery. We
    # reraise the exception as an AssertionError so that pytest will show it
    # in a semi-reasonable way

    try:
        subprocess.run([sys.executable, '-c', f'''\
from ndindex.tests.test_no_dependencies import {func}

{func}("numpy")'''], check=True, capture_output=True, encoding='utf-8')
    except subprocess.CalledProcessError as e:
        print(e.stdout, end='')
        raise AssertionError(e.stderr)

@pytest.mark.no_cover
@pytest.mark.parametrize('func', [func for func in globals() if func.startswith('_test')])
def test_sympy_dependency(func):
    # This unfortunately won't go through any of the pytest machinery. We
    # reraise the exception as an AssertionError so that pytest will show it
    # in a semi-reasonable way

    try:
        subprocess.run([sys.executable, '-c', f'''\
from ndindex.tests.test_no_dependencies import {func}

{func}("sympy")'''], check=True, capture_output=True, encoding='utf-8')
    except subprocess.CalledProcessError as e:
        print(e.stdout, end='')
        raise AssertionError(e.stderr)



import inspect
import warnings
import copy
import pickle

import numpy as np

from hypothesis import given, example, settings
from hypothesis.strategies import sampled_from

from pytest import raises

from ..array import ArrayIndex
from ..ndindex import ndindex
from ..booleanarray import BooleanArray
from ..integer import Integer
from ..ellipsis import ellipsis
from ..integerarray import IntegerArray
from ..slice import Slice
from ..tuple import Tuple
from .helpers import ndindices, check_same, assert_equal

@example(None)
@example([1, 2])
@given(ndindices)
def test_eq(idx):
    index = ndindex(idx)
    new = type(index)(*index.args)

    assert_equal(new.raw, index.raw)

    def check_eq(a, b):
        assert a == b
        assert b == a
        assert not (a != b)
        assert not (b != a)

    def check_neq(a, b):
        assert a != b
        assert b != a
        assert not (a == b)
        assert not (b == a)

    check_eq(new, index)
    check_eq(new.raw, index)
    check_eq(new, index.raw)

    check_eq(index.raw, index)
    assert hash(new) == hash(index)
    check_neq(index, 'a')

    try:
        h = hash(idx)
    except TypeError:
        pass
    else:
        assert hash(index) == h

    try:
        h = hash(index.raw)
    except TypeError:
        pass
    else:
        assert hash(index) == h

def test_eq_array_raises():
    index = ndindex([1, 2, 3])
    with raises(TypeError):
        np.equal(index.raw, index)
    with raises(TypeError):
        np.array_equal(index.raw, index)

def test_eq_explicit():
    assert Integer(0) != False
    assert Integer(1) != True
    assert Integer(0) != IntegerArray(0)
    assert IntegerArray([0, 1]) != [False, True]
    assert IntegerArray([0, 1]) == [0, 1]
    assert BooleanArray([False, True]) != [0, 1]
    assert BooleanArray([False, True]) == [False, True]

@example((np.array([1, 2]), 0))
@example([1, 2, 3])
@given(ndindices)
def test_ndindex(idx):
    index = ndindex(idx)
    assert index == idx
    assert ndindex[idx] == index

    def assert_raw_eq(idx, index):
        if isinstance(idx, (list, bool, np.bool_)):
            assert isinstance(index, ArrayIndex)
            assert index.dtype in [np.intp, np.bool_]
            assert_equal(index.raw, np.asarray(idx, dtype=index.dtype))
        elif isinstance(idx, np.integer):
            assert type(index) is Integer
            assert_equal(index.raw, int(idx))
        elif isinstance(idx, tuple):
            assert type(index.raw) is tuple
            assert len(idx) == len(index.raw)
            assert index.args == index.raw
            for i, j in zip(index.args, idx):
                assert_raw_eq(j, i)
        else:
            assert_equal(index.raw, idx)

    assert_raw_eq(idx, index)
    assert ndindex(index.raw) == index

def test_ndindex_invalid():
    a = np.arange(10)
    for idx in [1.0, [1.0], np.array([1.0]), np.array([1], dtype=object),
                np.array([])]:
        check_same(a, idx)

    # Older versions of NumPy gives a deprecation warning for this index. We
    # are not going to allow indices that give deprecation warnings in
    # ndindex.
    with warnings.catch_warnings(record=True) as r:
        # Make sure no warnings are emitted from ndindex()
        warnings.simplefilter("error")
        # Newer numpy versions raise ValueError with this index (although
        # perhaps they shouldn't)
        raises((IndexError, ValueError), lambda: ndindex([1, []]))
    assert not r

def test_ndindex_ellipsis():
    raises(IndexError, lambda: ndindex(ellipsis))

def test_signature():
    sig = inspect.signature(Integer)
    assert sig.parameters.keys() == {'idx'}


@example(([0, 1],))
@example((IntegerArray([], (0, 1)),))
@example(IntegerArray([], (0, 1)))
@example((1, ..., slice(1, 2)))
# eval can sometimes be slower than the default deadline of 200ms for large
# array indices
@settings(deadline=None)
@given(ndindices)
def test_repr_str(idx):
    # The repr form should be re-creatable
    index = ndindex(idx)
    d = {}
    exec("from ndindex import *", d)
    assert eval(repr(index), d) == idx

    # Str may not be re-creatable. Just test that it doesn't give an exception.
    str(index)

# _Tuple does not serialize properly with protocols 0 and 1. Support could
# probably be added if this is necessary.
LOWEST_SUPPORTED_PROTOCOL = 2
protocols = ["copy", "deepcopy"] + list(range(LOWEST_SUPPORTED_PROTOCOL, pickle.HIGHEST_PROTOCOL + 1))

@given(ndindices, sampled_from(protocols))
def test_serialization(idx, protocol):
    index = ndindex(idx)

    def serialize(index):
        if protocol == "copy":
            return copy.copy(index)
        elif protocol == "deepcopy":
            return copy.deepcopy(index)
        else:
            return pickle.loads(pickle.dumps(index, protocol=protocol))

    roundtripped = serialize(index)
    assert type(roundtripped) is type(index)
    assert roundtripped == index
    assert_equal(roundtripped.raw, index.raw)
    assert_equal(roundtripped.args, index.args)

    if isinstance(index, Slice):
        assert index._reduced == roundtripped._reduced == False
        s = index.reduce()
        assert s._reduced == True
        roundtripped_s = serialize(s)
        assert roundtripped_s._reduced == True

    if isinstance(index, Tuple):
        assert all([i._reduced == False for i in index.args if isinstance(i, Slice)])
        t = index.reduce()
        assert all([i._reduced == True for i in t.args if isinstance(i, Slice)])
        roundtripped_t = serialize(t)
        assert all([i._reduced == True for i in roundtripped_t.args if isinstance(i, Slice)])



import numpy as np
try:
    from numpy import AxisError as np_AxisError
except ImportError: # pragma: no cover
    from numpy.exceptions import AxisError as np_AxisError

from hypothesis import assume, given, example
from hypothesis.strategies import (one_of, integers, tuples as
                                   hypothesis_tuples, just, lists, shared,
                                   )

from pytest import raises

from ..ndindex import ndindex
from ..shapetools import (asshape, iter_indices, ncycles, BroadcastError,
                          AxisError, broadcast_shapes, remove_indices,
                          unremove_indices, associated_axis,
                          normalize_skip_axes)
from ..integer import Integer
from ..tuple import Tuple
from .helpers import (prod, mutually_broadcastable_shapes_with_skipped_axes,
                      skip_axes_st, mutually_broadcastable_shapes, tuples,
                      shapes, assert_equal, cross_shapes, cross_skip_axes,
                      cross_arrays_st, matmul_shapes, matmul_skip_axes,
                      matmul_arrays_st)

@example([[(1, 1), (1, 1)], (1,)], (0,))
@example([[(0,), (0,)], ()], (0,))
@example([[(1, 2), (2, 1)], (2,)], 1)
@given(mutually_broadcastable_shapes_with_skipped_axes, skip_axes_st)
def test_iter_indices(broadcastable_shapes, skip_axes):
    # broadcasted_shape will contain None on the skip_axes, as those axes
    # might not be broadcast compatible
    shapes, broadcasted_shape = broadcastable_shapes
    # We need no more than 31 dimensions so that the np.stack call below
    # doesn't fail.
    assume(len(broadcasted_shape) < 32)

    # 1. Normalize inputs
    _skip_axes = normalize_skip_axes(shapes, skip_axes)
    _skip_axes_kwarg_default = [()]*len(shapes)

    # Skipped axes may not be broadcast compatible. Since the index for a
    # skipped axis should always be a slice(None), the result should be the
    # same if the skipped axes are all moved to the end of the shape.
    canonical_shapes = []
    for s, sk in zip(shapes, _skip_axes):
        c = remove_indices(s, sk)
        canonical_shapes.append(c)

    non_skip_shapes = [remove_indices(shape, sk) for shape, sk in zip(shapes, _skip_axes)]
    assert np.broadcast_shapes(*non_skip_shapes) == broadcasted_shape

    nitems = prod(broadcasted_shape)

    if skip_axes == (): # kwarg default
        res = iter_indices(*shapes)
    else:
        res = iter_indices(*shapes, skip_axes=skip_axes)
    broadcasted_res = iter_indices(broadcasted_shape)

    sizes = [prod(shape) for shape in shapes]
    arrays = [np.arange(size).reshape(shape) for size, shape in zip(sizes, shapes)]
    canonical_sizes = [prod(shape) for shape in canonical_shapes]
    canonical_arrays = [np.arange(size).reshape(shape) for size, shape in zip(canonical_sizes, canonical_shapes)]
    canonical_broadcasted_array = np.arange(nitems).reshape(broadcasted_shape)

    # 2. Check that iter_indices is the same whether or not the shapes are
    # broadcasted together first. Also check that every iterated index is the
    # expected type and there are as many as expected.
    vals = []
    bvals = []
    n = -1

    def _remove_slices(idx):
        assert isinstance(idx, Tuple)
        idx2 = [i for i in idx.args if i != slice(None)]
        return Tuple(*idx2)

    for n, (idxes, bidxes) in enumerate(zip(res, broadcasted_res)):
        assert len(idxes) == len(shapes)
        assert len(bidxes) == 1
        for idx, shape, sk in zip(idxes, shapes, _skip_axes):
            assert isinstance(idx, Tuple)
            assert len(idx.args) == len(shape)

            for i in range(-1, -len(idx.args) - 1, -1):
                if i in sk:
                    assert idx.args[i] == slice(None)
                else:
                    assert isinstance(idx.args[i], Integer)

        canonical_idxes = [_remove_slices(idx) for idx in idxes]
        a_indexed = tuple([a[idx.raw] for a, idx in zip(arrays, idxes)])
        canonical_a_indexed = tuple([a[idx.raw] for a, idx in
                                  zip(canonical_arrays, canonical_idxes)])
        canonical_b_indexed = canonical_broadcasted_array[bidxes[0].raw]

        for c_indexed in canonical_a_indexed:
            assert c_indexed.shape == ()
        assert canonical_b_indexed.shape == ()

        if _skip_axes != _skip_axes_kwarg_default:
            vals.append(tuple(canonical_a_indexed))
        else:
            vals.append(a_indexed)

        bvals.append(canonical_b_indexed)

    # assert both iterators have the same length
    raises(StopIteration, lambda: next(res))
    raises(StopIteration, lambda: next(broadcasted_res))

    # Check that the correct number of items are iterated
    assert n == nitems - 1
    assert len(set(vals)) == len(vals) == nitems

    # 3. Check that every element of the (broadcasted) arrays is represented
    # by an iterated index.

    # The indices should correspond to the values that would be matched up
    # if the arrays were broadcasted together.
    if not arrays:
        assert vals == [()]
    else:
        correct_vals = list(zip(*[x.flat for x in np.broadcast_arrays(*canonical_arrays)]))
        # Also test that the indices are produced in a lexicographic order
        # (even though this isn't strictly guaranteed by the iter_indices
        # docstring) in the case when there are no skip axes. The order when
        # there are skip axes is more complicated because the skipped axes are
        # iterated together.
        if _skip_axes == _skip_axes_kwarg_default:
            assert vals == correct_vals
        else:
            assert set(vals) == set(correct_vals)
        assert bvals == list(canonical_broadcasted_array.flat)

@given(cross_arrays_st(), cross_shapes, cross_skip_axes)
def test_iter_indices_cross(cross_arrays, broadcastable_shapes, _skip_axes):
    # Test iter_indices behavior against np.cross, which effectively skips the
    # crossed axis. Note that we don't test against cross products of size 2
    # because a 2 x 2 cross product just returns the z-axis (i.e., it doesn't
    # actually skip an axis in the result shape), and also that behavior is
    # going to be removed in NumPy 2.0.
    a, b = cross_arrays
    shapes, broadcasted_shape = broadcastable_shapes

    # Sanity check
    skip_axes = normalize_skip_axes([*shapes, broadcasted_shape], _skip_axes)
    for sh, sk in zip([*shapes, broadcasted_shape], skip_axes):
        assert len(sk) == 1
        assert sh[sk[0]] == 3

    res = np.cross(a, b, axisa=skip_axes[0][0], axisb=skip_axes[1][0], axisc=skip_axes[2][0])
    assert res.shape == broadcasted_shape

    for idx1, idx2, idx3 in iter_indices(*shapes, broadcasted_shape, skip_axes=_skip_axes):
        assert a[idx1.raw].shape == (3,)
        assert b[idx2.raw].shape == (3,)
        assert_equal(np.cross(
            a[idx1.raw],
            b[idx2.raw]),
                     res[idx3.raw])

@given(matmul_arrays_st(), matmul_shapes, matmul_skip_axes)
def test_iter_indices_matmul(matmul_arrays, broadcastable_shapes, skip_axes):
    # Test iter_indices behavior against np.matmul, which effectively skips the
    # contracted axis (they aren't broadcasted together, even when they are
    # broadcast compatible).
    a, b = matmul_arrays
    shapes, broadcasted_shape = broadcastable_shapes

    # Note, we don't use normalize_skip_axes here because it sorts the skip
    # axes

    ax1, ax2 = skip_axes[0]
    ax3 = skip_axes[1][1]
    n, m, k = shapes[0][ax1], shapes[0][ax2], shapes[1][ax3]

    # Sanity check
    sk0, sk1, sk2 = skip_axes
    shape1, shape2 = shapes
    assert a.shape == shape1
    assert b.shape == shape2
    assert shape1[sk0[0]] == n
    assert shape1[sk0[1]] == m
    assert shape2[sk1[0]] == m
    assert shape2[sk1[1]] == k
    assert broadcasted_shape[sk2[0]] == n
    assert broadcasted_shape[sk2[1]] == k

    res = np.matmul(a, b, axes=skip_axes)
    assert res.shape == broadcasted_shape

    is_ordered = lambda sk, shape: (Integer(sk[0]).reduce(len(shape)).raw <= Integer(sk[1]).reduce(len(shape)).raw)
    orders = [
        is_ordered(sk0, shapes[0]),
        is_ordered(sk1, shapes[1]),
        is_ordered(sk2, broadcasted_shape),
    ]

    for idx1, idx2, idx3 in iter_indices(*shapes, broadcasted_shape, skip_axes=skip_axes):
        assert a[idx1.raw].shape == (n, m) if orders[0] else (m, n)
        assert b[idx2.raw].shape == (m, k) if orders[1] else (k, m)
        sub_res_axes = [
            (0, 1) if orders[0] else (1, 0),
            (0, 1) if orders[1] else (1, 0),
            (0, 1) if orders[2] else (1, 0),
        ]
        sub_res = np.matmul(a[idx1.raw], b[idx2.raw], axes=sub_res_axes)
        assert_equal(sub_res, res[idx3.raw])

def test_iter_indices_errors():
    try:
        list(iter_indices((10,), skip_axes=(2,)))
    except AxisError as e:
        ndindex_msg = str(e)
    else:
        raise RuntimeError("iter_indices did not raise AxisError") # pragma: no cover

    # Check that the message is the same one used by NumPy
    try:
        np.sum(np.arange(10), axis=2)
    except np_AxisError as e:
        np_msg = str(e)
    else:
        raise RuntimeError("np.sum() did not raise AxisError") # pragma: no cover

    assert ndindex_msg == np_msg

    try:
        list(iter_indices((2, 3), (3, 2)))
    except BroadcastError as e:
        ndindex_msg = str(e)
    else:
        raise RuntimeError("iter_indices did not raise BroadcastError") # pragma: no cover

    try:
        np.broadcast_shapes((2, 3), (3, 2))
    except ValueError as e:
        np_msg = str(e)
    else:
        raise RuntimeError("np.broadcast_shapes() did not raise ValueError") # pragma: no cover


    if 'Mismatch' in str(np_msg): # pragma: no cover
        # Older versions of NumPy do not have the more helpful error message
        assert ndindex_msg == np_msg

    with raises(ValueError, match=r"not unique"):
        list(iter_indices((1, 2), skip_axes=(0, 1, 0)))

    raises(AxisError, lambda: list(iter_indices((0,), skip_axes=(3,))))
    raises(ValueError, lambda: list(iter_indices(skip_axes=(0,))))
    raises(TypeError, lambda: list(iter_indices(1, 2)))
    raises(TypeError, lambda: list(iter_indices(1, 2, (2, 2))))
    raises(TypeError, lambda: list(iter_indices([(1, 2), (2, 2)])))

@example(1, 1, 1)
@given(integers(0, 100), integers(0, 100), integers(0, 100))
def test_ncycles(i, n, m):
    N = ncycles(range(i), n)
    if n == 1:
        assert N == range(i)
    else:
        assert isinstance(N, ncycles)
        assert N.iterable == range(i)
        assert N.n == n
        assert f"range(0, {i})" in repr(N)
        assert str(n) in repr(N)

    L = list(N)
    assert len(L) == i*n
    for j in range(i*n):
        assert L[j] == j % i

    M = ncycles(N, m)
    if n*m == 1:
        assert M == range(i)
    else:
        assert isinstance(M, ncycles)
        assert M.iterable == range(i)
        assert M.n == n*m

@given(one_of(mutually_broadcastable_shapes,
              hypothesis_tuples(tuples(shapes), just(None))))
def test_broadcast_shapes(broadcastable_shapes):
    shapes, broadcasted_shape = broadcastable_shapes
    if broadcasted_shape is not None:
        assert broadcast_shapes(*shapes) == broadcasted_shape

    arrays = [np.empty(shape) for shape in shapes]
    broadcastable = True
    try:
        broadcasted_shape = np.broadcast(*arrays).shape
    except ValueError:
        broadcastable = False

    if broadcastable:
        assert broadcast_shapes(*shapes) == broadcasted_shape
    else:
        raises(BroadcastError, lambda: broadcast_shapes(*shapes))


@given(lists(shapes, max_size=32))
def test_broadcast_shapes_errors(shapes):
    error = True
    try:
        broadcast_shapes(*shapes)
    except BroadcastError as exc:
        e = exc
    else:
        error = False

    # The ndindex and numpy errors won't match in general, because
    # ndindex.broadcast_shapes gives an error with the first two shapes that
    # aren't broadcast compatible, but numpy doesn't always, due to different
    # implementation algorithms (e.g., the message from
    # np.broadcast_shapes((0,), (0, 2), (2, 0)) mentions the last two shapes
    # whereas ndindex.broadcast_shapes mentions the first two).

    # Instead, just confirm that the error message is correct as stated, and
    # check against the numpy error message when just broadcasting the two
    # reportedly bad shapes.

    if not error:
        try:
            np.broadcast_shapes(*shapes)
        except: # pragma: no cover
            raise RuntimeError("ndindex.broadcast_shapes raised but np.broadcast_shapes did not")
        return

    assert shapes[e.arg1] == e.shape1
    assert shapes[e.arg2] == e.shape2

    try:
        np.broadcast_shapes(e.shape1, e.shape2)
    except ValueError as np_exc:
        # Check that they do in fact not broadcast, and the error messages are
        # the same modulo the different arg positions.
        if 'Mismatch' in str(np_exc): # pragma: no cover
            # Older versions of NumPy do not have the more helpful error message
            assert str(BroadcastError(0, e.shape1, 1, e.shape2)) == str(np_exc)
    else: # pragma: no cover
        raise RuntimeError("ndindex.broadcast_shapes raised but np.broadcast_shapes did not")

    raises(TypeError, lambda: broadcast_shapes(1, 2))
    raises(TypeError, lambda: broadcast_shapes(1, 2, (2, 2)))
    raises(TypeError, lambda: broadcast_shapes([(1, 2), (2, 2)]))

@given(mutually_broadcastable_shapes_with_skipped_axes, skip_axes_st)
def test_broadcast_shapes_skip_axes(broadcastable_shapes, skip_axes):
    shapes, broadcasted_shape = broadcastable_shapes
    assert broadcast_shapes(*shapes, skip_axes=skip_axes) == broadcasted_shape

@example([[], ()], (0,))
@example([[(0, 1)], (0, 1)], (2,))
@example([[(0, 1)], (0, 1)], (0, -1))
@example([[(0, 1, 0, 0, 0), (2, 0, 0, 0)], (0, 2, 0, 0, 0)], [1])
@given(mutually_broadcastable_shapes,
       one_of(
           integers(-20, 20),
           tuples(integers(-20, 20), max_size=20),
           lists(tuples(integers(-20, 20), max_size=20), max_size=32)))
def test_broadcast_shapes_skip_axes_errors(broadcastable_shapes, skip_axes):
    shapes, broadcasted_shape = broadcastable_shapes

    # All errors should come from normalize_skip_axes, which is tested
    # separately below.
    try:
        normalize_skip_axes(shapes, skip_axes)
    except (TypeError, ValueError, IndexError) as e:
        raises(type(e), lambda: broadcast_shapes(*shapes,
                                                   skip_axes=skip_axes))
        return

    try:
        broadcast_shapes(*shapes, skip_axes=skip_axes)
    except IndexError:
        raise RuntimeError("broadcast_shapes raised but should not have") # pragma: no cover
    except BroadcastError:
        # Broadcastable shapes can become unbroadcastable after skipping axes
        # (see the @example above).
        pass

remove_indices_n = shared(integers(0, 100))

@given(remove_indices_n,
       remove_indices_n.flatmap(lambda n: lists(integers(-n, n), unique=True)))
def test_remove_indices(n, idxes):
    if idxes:
        assume(max(idxes) < n)
        assume(min(idxes) >= -n)
    a = tuple(range(n))
    b = remove_indices(a, idxes)
    if len(idxes) == 1:
        assert remove_indices(a, idxes[0]) == b

    A = list(a)
    for i in idxes:
        A[i] = None

    assert set(A) - set(b) == ({None} if idxes else set())
    assert set(b) - set(A) == set()

    # Check the order is correct
    j = 0
    for i in range(n):
        val = A[i]
        if val == None:
            assert val not in b
        else:
            assert b[j] == val
            j += 1

    # Test that unremove_indices is the inverse
    if all(i >= 0 for i in idxes) or all(i < 0 for i in idxes):
        assert unremove_indices(b, idxes) == tuple(A)
    else:
        raises(NotImplementedError, lambda: unremove_indices(b, idxes))

# Meta-test for the hypothesis strategy
@given(mutually_broadcastable_shapes_with_skipped_axes, skip_axes_st)
def test_mutually_broadcastable_shapes_with_skipped_axes(broadcastable_shapes,
                                                         skip_axes): # pragma: no cover
    shapes, broadcasted_shape = broadcastable_shapes
    _skip_axes = normalize_skip_axes(shapes, skip_axes)

    assert len(_skip_axes) == len(shapes)

    for shape in shapes:
        assert None not in shape
    assert None not in broadcasted_shape

    _shapes = [remove_indices(shape, sk) for shape, sk in zip(shapes, _skip_axes)]

    assert broadcast_shapes(*_shapes) == broadcasted_shape

@example([[(2, 10, 3, 4), (10, 3, 4)], (2, 3, 4)], (-3,))
@example([[(0, 10, 2, 3, 10, 4), (1, 10, 1, 0, 10, 2, 3, 4)],
          (1, 1, 0, 2, 3, 4)], (1, 4))
@example([[(2, 0, 3, 4)], (2, 3, 4)], (1,))
@example([[(0, 0, 0, 0, 0), (0, 0, 0, 0, 0, 0)], (0, 0, 0, 0)], (1, 2))
@given(mutually_broadcastable_shapes_with_skipped_axes, skip_axes_st)
def test_associated_axis(broadcastable_shapes, skip_axes):
    shapes, broadcasted_shape = broadcastable_shapes
    _skip_axes = normalize_skip_axes(shapes, skip_axes)

    for shape, sk in zip(shapes, _skip_axes):
        n = len(shape)
        for i in range(-len(shape), 0):
            val = shape[i]

            bval = associated_axis(broadcasted_shape, i, sk)
            if bval is None:
                assert ndindex(i).reduce(n, negative_int=True) in sk, (shape, i)
            else:
                assert val == 1 or bval == val, (shape, i)


    sk = max(_skip_axes, key=len, default=())
    for i in range(-len(broadcasted_shape)-len(sk)-10, -len(broadcasted_shape)-len(sk)):
        assert associated_axis(broadcasted_shape, i, sk) is None

# TODO: add a hypothesis test for asshape
def test_asshape():
    assert asshape(1) == (1,)
    assert asshape(np.int64(2)) == (2,)
    assert type(asshape(np.int64(2))[0]) == int
    assert asshape((1, 2)) == (1, 2)
    assert asshape([1, 2]) == (1, 2)
    assert asshape((1, 2), allow_int=False) == (1, 2)
    assert asshape([1, 2], allow_int=False) == (1, 2)
    assert asshape((np.int64(1), np.int64(2))) == (1, 2)
    assert type(asshape((np.int64(1), np.int64(2)))[0]) == int
    assert type(asshape((np.int64(1), np.int64(2)))[1]) == int
    assert asshape((-1, -2), allow_negative=True) == (-1, -2)
    assert asshape(-2, allow_negative=True) == (-2,)


    raises(TypeError, lambda: asshape(1.0))
    raises(TypeError, lambda: asshape((1.0,)))
    raises(ValueError, lambda: asshape(-1))
    raises(ValueError, lambda: asshape((1, -1)))
    raises(ValueError, lambda: asshape((1, None)))
    raises(TypeError, lambda: asshape(...))
    raises(TypeError, lambda: asshape(Integer(1)))
    raises(TypeError, lambda: asshape(Tuple(1, 2)))
    raises(TypeError, lambda: asshape((True,)))
    raises(TypeError, lambda: asshape({1, 2}))
    raises(TypeError, lambda: asshape({1: 2}))
    raises(TypeError, lambda: asshape('1'))
    raises(TypeError, lambda: asshape(1, allow_int=False))
    raises(TypeError, lambda: asshape(-1, allow_int=False))
    raises(TypeError, lambda: asshape(-1, allow_negative=True, allow_int=False))
    raises(TypeError, lambda: asshape(np.int64(1), allow_int=False))
    raises(IndexError, lambda: asshape((2, 3), 3))

@example([(0,), ()], (0, 0))
@example([(0, 1), (0,), ()], [(-1,), (0,), ()])
@example([(5,)], (10,))
@example([], [])
@example([()], [])
@example([(0, 1)], 0)
@example([(2, 3), (2, 3, 4)], [(3,), (0,)])
@example([(0, 1)], 0)
@example([(2, 3)], (0, -2))
@example([(2, 4), (2, 3, 4)], [(0,), (-3,)])
@example([(1, 2)], [(0,), (1,)])
@given(lists(tuples(integers(0))),
       one_of(integers(), tuples(integers()), lists(tuples(integers()))))
def test_normalize_skip_axes(shapes, skip_axes):
    if not shapes:
        if skip_axes in [(), []]:
            assert normalize_skip_axes(shapes, skip_axes) == []
        else:
            raises(ValueError, lambda: normalize_skip_axes(shapes, skip_axes))
        return

    min_dim = min(len(shape) for shape in shapes)

    if isinstance(skip_axes, int):
        if not (-min_dim <= skip_axes < min_dim):
            raises(AxisError, lambda: normalize_skip_axes(shapes, skip_axes))
            return
        _skip_axes = [(skip_axes,)]*len(shapes)
    elif isinstance(skip_axes, tuple):
        if not all(-min_dim <= s < min_dim for s in skip_axes):
            raises(AxisError, lambda: normalize_skip_axes(shapes, skip_axes))
            return
        _skip_axes = [skip_axes]*len(shapes)
    elif not skip_axes:
        # empty list will be interpreted as a single skip_axes tuple
        assert normalize_skip_axes(shapes, skip_axes) == [()]*len(shapes)
        return
    else:
        if len(shapes) != len(skip_axes):
            raises(ValueError, lambda: normalize_skip_axes(shapes, skip_axes))
            return
        _skip_axes = skip_axes

    try:
        res = normalize_skip_axes(shapes, skip_axes)
    except AxisError as e:
        axis, ndim = e.args
        assert any(axis in s for s in _skip_axes)
        assert any(ndim == len(shape) for shape in shapes)
        assert axis < -ndim or axis >= ndim
        return
    except ValueError as e:
        if 'not unique' in str(e):
            bad_skip_axes, bad_shape = e.skip_axes, e.shape
            assert str(bad_skip_axes) in str(e)
            assert str(bad_shape) in str(e)
            assert bad_skip_axes in _skip_axes
            assert bad_shape in shapes
            indexed = [bad_shape[i] for i in bad_skip_axes]
            assert len(indexed) != len(set(indexed))
            return
        else: # pragma: no cover
            raise

    assert isinstance(res, list)
    assert all(isinstance(x, tuple) for x in res)
    assert all(isinstance(i, int) for x in res for i in x)

    assert len(res) == len(shapes)
    for shape, new_skip_axes in zip(shapes, res):
        assert len(new_skip_axes) == len(set(new_skip_axes))
        assert new_skip_axes == tuple(sorted(new_skip_axes))
        for i in new_skip_axes:
            assert i < 0
            assert ndindex(i).reduce(len(shape), negative_int=True) == i

    # TODO: Assert the order is maintained (doesn't actually matter for now
    # but could for future applications)

def test_normalize_skip_axes_errors():
    raises(TypeError, lambda: normalize_skip_axes([(1,)], {0: 1}))
    raises(TypeError, lambda: normalize_skip_axes([(1,)], {0}))
    raises(TypeError, lambda: normalize_skip_axes([(1,)], [(0,), 0]))
    raises(TypeError, lambda: normalize_skip_axes([(1,)], [0, (0,)]))

@example(10, 5)
@given(integers(), integers())
def test_axiserror(axis, ndim):
    if ndim == 0 and axis in [0, -1]:
        # NumPy allows axis=0 or -1 for 0-d arrays
        AxisError(axis, ndim)
        return

    try:
        if ndim >= 0:
            range(ndim)[axis]
    except IndexError:
        e = AxisError(axis, ndim)
    else:
        raises(ValueError, lambda: AxisError(axis, ndim))
        return

    try:
        raise e
    except AxisError as e2:
        assert e2.args == (axis, ndim)
        if ndim <= 32 and -1000 < axis < 1000:
            a = np.empty((0,)*ndim)
            try:
                np.sum(a, axis=axis)
            except np_AxisError as e3:
                assert str(e2) == str(e3)
            else:
                raise RuntimeError("numpy didn't raise AxisError") # pragma: no cover



from math import gcd

from .._crt import crt, ilcm, gcdex

from hypothesis import given, example
from hypothesis.strategies import integers, lists, shared

from sympy.ntheory.modular import crt as crt_sympy
size = shared(integers(min_value=1, max_value=10))
@given(
    size.flatmap(lambda s: lists(integers(min_value=1), min_size=s, max_size=s)),
    size.flatmap(lambda s: lists(integers(), min_size=s, max_size=s)),
)
def test_crt(m, v):
    res = crt(m, v)

    if res is not None:
        for m_i, v_i in zip(m, v):
            assert v_i % m_i == res % m_i

        assert res == crt_sympy(m, v)[0]
    else:
        assert crt_sympy(m, v) is None

@example(1, 2)
@given(integers(min_value=0), integers(min_value=0))
def test_ilcm(x, y):
    L = ilcm(x, y)

    if 0 in [x, y]:
        assert L == 0
        return

    assert L >= x
    assert L >= y

    # L is a common multiple
    assert L % x == 0
    assert L % y == 0

    if L - min(x, y) <= 1000:
        # L is the least common multiple
        for i in range(min(x, y), L):
            assert i % x != 0 or i % y != 0

@example(0, 3)
@example(3, 0)
@example(0, 0)
@given(integers(), integers())
def test_gcdex(a, b):
    x, y, g = gcdex(a, b)

    assert g == gcd(a, b)
    assert x*a + y*b == g



import sys
from itertools import chain
import warnings
from functools import wraps, partial

from numpy import ndarray, generic, intp, bool_, asarray, broadcast_shapes
import numpy.testing

from pytest import fail

from hypothesis import assume, note
from hypothesis.strategies import (integers, none, one_of, lists, just,
                                   builds, shared, composite, sampled_from,
                                   nothing, tuples as hypothesis_tuples)
from hypothesis.extra.numpy import (arrays, mutually_broadcastable_shapes as
                                    mbs, BroadcastableShapes, valid_tuple_axes)
from hypothesis.errors import UnsatisfiedAssumption

from ..ndindex import ndindex
from ..shapetools import remove_indices, unremove_indices
from .._crt import prod

# Hypothesis strategies for generating indices. Note that some of these
# strategies are nominally already defined in hypothesis, but we redefine them
# here because the hypothesis definitions are too restrictive. For example,
# hypothesis's slices strategy does not generate slices with negative indices.
# Similarly, hypothesis.extra.numpy.basic_indices only generates tuples.

nonnegative_ints = integers(0, 10)
negative_ints = integers(-10, -1)
ints = lambda: one_of(nonnegative_ints, negative_ints)

def slices(start=one_of(none(), ints()), stop=one_of(none(), ints()),
           step=one_of(none(), ints())):
    return builds(slice, start, stop, step)

ellipses = lambda: just(...)
newaxes = lambda: just(None)

# hypotheses.strategies.tuples only generates tuples of a fixed size
def tuples(elements, *, min_size=0, max_size=None, unique_by=None, unique=False):
    return lists(elements, min_size=min_size, max_size=max_size,
                 unique_by=unique_by, unique=unique).map(tuple)

MAX_ARRAY_SIZE = 100000
SHORT_MAX_ARRAY_SIZE = 1000
shapes = tuples(integers(0, 10)).filter(
             # numpy gives errors with empty arrays with large shapes.
             # See https://github.com/numpy/numpy/issues/15753
             lambda shape: prod([i for i in shape if i]) < MAX_ARRAY_SIZE)

_short_shapes = lambda n: tuples(integers(0, 10), min_size=n, max_size=32).filter(
             # numpy gives errors with empty arrays with large shapes.
             # See https://github.com/numpy/numpy/issues/15753
             lambda shape: prod([i for i in shape if i]) < SHORT_MAX_ARRAY_SIZE)

# short_shapes should be used in place of shapes in any test function that
# uses ndindices, boolean_arrays, or tuples
short_shapes = shared(_short_shapes(0))

_integer_arrays = arrays(intp, short_shapes)
integer_scalars = arrays(intp, ()).map(lambda x: x[()])
integer_arrays = one_of(integer_scalars, _integer_arrays.flatmap(lambda x: one_of(just(x), just(x.tolist()))))

# We need to make sure shapes for boolean arrays are generated in a way that
# makes them related to the test array shape. Otherwise, it will be very
# difficult for the boolean array index to match along the test array, which
# means we won't test any behavior other than IndexError.

@composite
def subsequences(draw, sequence):
    seq = draw(sequence)
    start = draw(integers(0, max(0, len(seq)-1)))
    stop = draw(integers(start, len(seq)))
    return seq[start:stop]

_boolean_arrays = arrays(bool_, one_of(subsequences(short_shapes), short_shapes))
boolean_scalars = arrays(bool_, ()).map(lambda x: x[()])
boolean_arrays = one_of(boolean_scalars, _boolean_arrays.flatmap(lambda x: one_of(just(x), just(x.tolist()))))

def _doesnt_raise(idx):
    try:
        ndindex(idx)
    except (IndexError, ValueError, NotImplementedError):
        return False
    return True

Tuples = tuples(one_of(ellipses(), ints(), slices(), newaxes(),
                       integer_arrays, boolean_arrays)).filter(_doesnt_raise)

ndindices = one_of(
    ints(),
    slices(),
    ellipses(),
    newaxes(),
    Tuples,
    integer_arrays,
    boolean_arrays,
).filter(_doesnt_raise)

# Note: We could use something like this:

# mutually_broadcastable_shapes = shared(integers(1, 32).flatmap(lambda i: mbs(num_shapes=i).filter(
#     lambda broadcastable_shapes: prod([i for i in broadcastable_shapes.result_shape if i]) < MAX_ARRAY_SIZE)))

@composite
def _mutually_broadcastable_shapes(draw, *, shapes=short_shapes, min_shapes=0, max_shapes=32, min_side=0):
    # mutually_broadcastable_shapes() with the default inputs doesn't generate
    # very interesting examples (see
    # https://github.com/HypothesisWorks/hypothesis/issues/3170). It's very
    # difficult to get it to do so by tweaking the max_* parameters, because
    # making them too big leads to generating too large shapes and filtering
    # too much. So instead, we trick it into generating more interesting
    # examples by telling it to create shapes that broadcast against some base
    # shape.

    # Unfortunately, this, along with the filtering below, has a downside that
    # it tends to generate a result shape of () more often than you might
    # like. But it generates enough "real" interesting shapes that both of
    # these workarounds are worth doing (plus I don't know if any other better
    # way of handling the situation).
    base_shape = draw(shapes)

    input_shapes, result_shape = draw(
        mbs(
            num_shapes=max_shapes,
            base_shape=base_shape,
            min_side=min_side,
        ))

    # The hypothesis mutually_broadcastable_shapes doesn't allow num_shapes to
    # be a strategy. It's tempting to do something like num_shapes =
    # draw(integers(min_shapes, max_shapes)), but this shrinks poorly. See
    # https://github.com/HypothesisWorks/hypothesis/issues/3151. So instead of
    # using a strategy to draw the number of shapes, we just generate max_shapes
    # shapes and pick a subset of them.
    final_input_shapes = draw(lists(sampled_from(input_shapes),
                                    min_size=min_shapes, max_size=max_shapes))


    # Note: result_shape is input_shapes broadcasted with base_shape, but
    # base_shape itself is not part of input_shapes. We "really" want our base
    # shape to be (). We are only using it here to trick
    # mutually_broadcastable_shapes into giving more interesting examples.
    final_result_shape = broadcast_shapes(*final_input_shapes)

    # The broadcast compatible shapes can be bigger than the base shape. This
    # is already somewhat limited by the mutually_broadcastable_shapes
    # defaults, and pretty unlikely, but we filter again here just to be safe.
    if not prod([i for i in final_result_shape if i]) < SHORT_MAX_ARRAY_SIZE: # pragma: no cover
        note(f"Filtering the shape {result_shape} (too many elements)")
        assume(False)

    return BroadcastableShapes(final_input_shapes, final_result_shape)

mutually_broadcastable_shapes = shared(_mutually_broadcastable_shapes())

def _fill_shape(draw,
                *,
                result_shape,
                skip_axes,
                skip_axes_values):
    max_n = max([i + 1 if i >= 0 else -i for i in skip_axes], default=0)
    assume(max_n <= len(skip_axes) + len(result_shape))
    dim = draw(integers(min_value=max_n, max_value=len(skip_axes) + len(result_shape)))
    new_shape = ['placeholder']*dim
    for i in skip_axes:
        assume(new_shape[i] is not None) # skip_axes must be unique
        new_shape[i] = None
    j = -1
    for i in range(-1, -dim - 1, -1):
        if new_shape[i] is None:
            new_shape[i] = draw(skip_axes_values)
        else:
            new_shape[i] = draw(sampled_from([result_shape[j], 1]))
            j -= 1
    while new_shape and new_shape[0] == 'placeholder': # pragma: no cover
        # Can happen if positive and negative skip_axes refer to the same
        # entry
        new_shape.pop(0)

    # This will happen if the skip axes are too large
    assume('placeholder' not in new_shape)

    if prod([i for i in new_shape if i]) >= SHORT_MAX_ARRAY_SIZE:
        note(f"Filtering the shape {new_shape} (too many elements)")
        assume(False)

    return tuple(new_shape)

skip_axes_with_broadcasted_shape_type = shared(sampled_from([int, tuple, list]))

@composite
def _mbs_and_skip_axes(
        draw,
        shapes=short_shapes,
        min_shapes=0,
        max_shapes=32,
        skip_axes_type_st=skip_axes_with_broadcasted_shape_type,
        skip_axes_values=integers(0, 20),
        num_skip_axes=None,
):
    """
    mutually_broadcastable_shapes except skip_axes() axes might not be
    broadcastable

    The result_shape will be None in the position of skip_axes.
    """
    skip_axes_type = draw(skip_axes_type_st)
    _result_shape = draw(shapes)
    if _result_shape == ():
        assume(num_skip_axes is None)

    ndim = len(_result_shape)
    num_shapes = draw(integers(min_value=min_shapes, max_value=max_shapes))
    if not num_shapes:
        assume(num_skip_axes is None)
        num_skip_axes = 0
    if not ndim:
        return BroadcastableShapes([()]*num_shapes, ()), ()

    if num_skip_axes is not None:
        min_skip_axes = max_skip_axes = num_skip_axes
    else:
        min_skip_axes = 0
        max_skip_axes = None

    # int and single tuple cases must be limited to N to ensure that they are
    # correct for all shapes
    if skip_axes_type == int:
        assume(num_skip_axes in [None, 1])
        skip_axes = draw(valid_tuple_axes(ndim, min_size=1, max_size=1))[0]
        _skip_axes = [(skip_axes,)]*num_shapes
    elif skip_axes_type == tuple:
        skip_axes = draw(tuples(integers(-ndim, ndim-1), min_size=min_skip_axes,
                               max_size=max_skip_axes, unique=True))
        _skip_axes = [skip_axes]*num_shapes
    elif skip_axes_type == list:
        skip_axes = []
        for i in range(num_shapes):
            skip_axes.append(draw(tuples(integers(-ndim, ndim+1), min_size=min_skip_axes,
                                         max_size=max_skip_axes, unique=True)))
        _skip_axes = skip_axes

    shapes = []
    for i in range(num_shapes):
        shapes.append(_fill_shape(draw, result_shape=_result_shape, skip_axes=_skip_axes[i],
                                  skip_axes_values=skip_axes_values))

    non_skip_shapes = [remove_indices(shape, sk) for shape, sk in
                            zip(shapes, _skip_axes)]
    # Broadcasting the result _fill_shape may produce a shape different from
    # _result_shape because it might not have filled all dimensions, or it
    # might have chosen 1 for a dimension every time. Ideally we would just be
    # using shapes from mutually_broadcastable_shapes, but I don't know how to
    # reverse inject skip axes into shapes in general (see the comment in
    # unremove_indices). So for now, we just use the actual broadcast of the
    # non-skip shapes. Note that we use np.broadcast_shapes here instead of
    # ndindex.broadcast_shapes because test_broadcast_shapes itself uses this
    # strategy.
    broadcasted_shape = broadcast_shapes(*non_skip_shapes)

    return BroadcastableShapes(shapes, broadcasted_shape), skip_axes

mbs_and_skip_axes = shared(_mbs_and_skip_axes())

mutually_broadcastable_shapes_with_skipped_axes = mbs_and_skip_axes.map(
    lambda i: i[0])
skip_axes_st = mbs_and_skip_axes.map(lambda i: i[1])

@composite
def _cross_shapes_and_skip_axes(draw):
    (shapes, _broadcasted_shape), skip_axes = draw(_mbs_and_skip_axes(
        shapes=_short_shapes(2),
        min_shapes=2,
        max_shapes=2,
        num_skip_axes=1,
        # TODO: Test other skip axes types
        skip_axes_type_st=just(list),
        skip_axes_values=just(3),
    ))

    broadcasted_skip_axis = draw(integers(-len(_broadcasted_shape)-1, len(_broadcasted_shape)))
    broadcasted_shape = unremove_indices(_broadcasted_shape,
                                         [broadcasted_skip_axis], val=3)
    skip_axes.append((broadcasted_skip_axis,))

    return BroadcastableShapes(shapes, broadcasted_shape), skip_axes

cross_shapes_and_skip_axes = shared(_cross_shapes_and_skip_axes())
cross_shapes = cross_shapes_and_skip_axes.map(lambda i: i[0])
cross_skip_axes = cross_shapes_and_skip_axes.map(lambda i: i[1])

@composite
def cross_arrays_st(draw):
    broadcastable_shapes = draw(cross_shapes)
    shapes, broadcasted_shape = broadcastable_shapes

    # Sanity check
    assert len(shapes) == 2
    # We need to generate fairly random arrays. Otherwise, if they are too
    # similar to each other, like two arange arrays would be, the cross
    # product will be 0. We also disable the fill feature in arrays() for the
    # same reason, as it would otherwise generate too many vectors that are
    # colinear.
    a = draw(arrays(dtype=int, shape=shapes[0], elements=integers(-100, 100), fill=nothing()))
    b = draw(arrays(dtype=int, shape=shapes[1], elements=integers(-100, 100), fill=nothing()))

    return a, b

@composite
def _matmul_shapes_and_skip_axes(draw):
    (shapes, _broadcasted_shape), skip_axes = draw(_mbs_and_skip_axes(
        shapes=_short_shapes(2),
        min_shapes=2,
        max_shapes=2,
        num_skip_axes=2,
        # TODO: Test other skip axes types
        skip_axes_type_st=just(list),
        skip_axes_values=just(None),
    ))

    broadcasted_skip_axes = draw(hypothesis_tuples(*[
        integers(-len(_broadcasted_shape)-1, len(_broadcasted_shape))
    ]*2))

    try:
        broadcasted_shape = unremove_indices(_broadcasted_shape,
                                             broadcasted_skip_axes)
    except NotImplementedError:
        # TODO: unremove_indices only works with both positive or both negative
        assume(False)
    # Make sure the indices are unique
    assume(len(set(broadcasted_skip_axes)) == len(broadcasted_skip_axes))

    skip_axes.append(broadcasted_skip_axes)

    # (n, m) @ (m, k) -> (n, k)
    n, m, k = draw(hypothesis_tuples(integers(0, 10), integers(0, 10),
                                     integers(0, 10)))
    shape1, shape2 = map(list, shapes)
    ax1, ax2 = skip_axes[0]
    shape1[ax1] = n
    shape1[ax2] = m
    ax1, ax2 = skip_axes[1]
    shape2[ax1] = m
    shape2[ax2] = k
    broadcasted_shape = list(broadcasted_shape)
    ax1, ax2 = skip_axes[2]
    broadcasted_shape[ax1] = n
    broadcasted_shape[ax2] = k

    shapes = (tuple(shape1), tuple(shape2))
    broadcasted_shape = tuple(broadcasted_shape)

    return BroadcastableShapes(shapes, broadcasted_shape), skip_axes

matmul_shapes_and_skip_axes = shared(_matmul_shapes_and_skip_axes())
matmul_shapes = matmul_shapes_and_skip_axes.map(lambda i: i[0])
matmul_skip_axes = matmul_shapes_and_skip_axes.map(lambda i: i[1])

@composite
def matmul_arrays_st(draw):
    broadcastable_shapes = draw(matmul_shapes)
    shapes, broadcasted_shape = broadcastable_shapes

    # Sanity check
    assert len(shapes) == 2
    a = draw(arrays(dtype=int, shape=shapes[0], elements=integers(-100, 100)))
    b = draw(arrays(dtype=int, shape=shapes[1], elements=integers(-100, 100)))

    return a, b

reduce_kwargs = sampled_from([{}, {'negative_int': False}, {'negative_int': True}])

def assert_equal(actual, desired, allow_scalar_0d=False, err_msg='', verbose=True):
    """
    Assert that two objects are equal.

    - If the objects are ndarrays, this is the same as
      numpy.testing.assert_equal except it also requires the shapes and dtypes
      to be equal

    - If the objects are tuples, recursively call assert_equal to support
      tuples of arrays.

    - If allow_scalar_0d=True, scalars will be considered equal to equivalent
      0-D arrays.

    - Require the types of actual and desired to be exactly the same
      (excepting for scalars when allow_scalar_0d=True).

    """
    if not (allow_scalar_0d and (isinstance(actual, generic)
                                 or isinstance(desired, generic))):
        assert type(actual) is type(desired), err_msg or f"{type(actual)} != {type(desired)}"

    if isinstance(actual, (ndarray, generic)):
        numpy.testing.assert_equal(actual, desired, err_msg=err_msg,
                                   verbose=verbose)
        assert actual.shape == desired.shape, err_msg or f"{actual.shape} != {desired.shape}"
        assert actual.dtype == desired.dtype, err_msg or f"{actual.dtype} != {desired.dtype}"
    elif isinstance(actual, tuple):
        assert len(actual) == len(desired), err_msg
        for i, j in zip(actual, desired):
            assert_equal(i, j, err_msg=err_msg, verbose=verbose)
    else:
        assert actual == desired, err_msg

assert_equal_allow_scalar_0d = partial(assert_equal, allow_scalar_0d=True)

def warnings_are_errors(f):
    @wraps(f)
    def inner(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            return f(*args, **kwargs)
    return inner

@warnings_are_errors
def check_same(a, idx, *, raw_func=lambda a, idx: a[idx],
               ndindex_func=lambda a, index: a[index.raw],
               conversion_func=ndindex,
               same_exception=True, assert_equal=assert_equal):
    """
    Check that a raw index idx produces the same result on an array a before
    and after being transformed by ndindex.

    Tests that raw_func(a, idx) == ndindex_func(a, ndindex(idx)) or that they
    raise the same exception. If same_exception=False, it will still check
    that they both raise an exception, but will not require the exception type
    and message to be the same.

    By default, raw_func(a, idx) is a[idx] and ndindex_func(a, index) is
    a[index.raw].

    The assert_equal argument changes the function used to test equality. By
    default it is the custom assert_equal() function in this file that extends
    numpy.testing.assert_equal. If the func functions return something other
    than arrays, assert_equal should be set to something else, like

        def assert_equal(x, y):
            assert x == y

    """
    exception = None
    try:
        # Handle list indices that NumPy treats as tuple indices with a
        # deprecation warning. We want to test against the post-deprecation
        # behavior.
        e_inner = None
        try:
            try:
                a_raw = raw_func(a, idx)
            except IndexError as e:
                # It's not straightforward to avoid indexing too many
                # dimensions in the strategy generation, because the total
                # number of dimensions in the result array is not a trivial
                # thing. Furthermore, some versions of NumPy limit this to 32
                # and some limit it to 64.
                if "number of dimensions must be within" in str(e): # pragma: no cover
                    raise UnsatisfiedAssumption
                raise
            except Warning as w:
                # In NumPy < 1.23, this is a FutureWarning. In 1.23 the
                # deprecation was removed and lists are always interpreted as
                # array indices.
                if ("Using a non-tuple sequence for multidimensional indexing is deprecated" in w.args[0]): # pragma: no cover
                    idx = asarray(idx, dtype=intp)
                    a_raw = raw_func(a, idx)
                elif "Out of bound index found. This was previously ignored when the indexing result contained no elements. In the future the index error will be raised. This error occurs either due to an empty slice, or if an array has zero elements even before indexing." in w.args[0]:
                    same_exception = False
                    raise IndexError
                else: # pragma: no cover
                    fail(f"Unexpected warning raised: {w}")
        except Exception:
            _, e_inner, _ = sys.exc_info()
        if e_inner:
            raise e_inner
    except UnsatisfiedAssumption: # pragma: no cover
        raise
    except Exception as e:
        exception = e

    index = '<conversion to ndindex object failed>'
    try:
        index = conversion_func(idx)
        a_ndindex = ndindex_func(a, index)
    except Exception as e:
        if not exception:
            fail(f"Raw form does not raise but ndindex form does ({e!r}): {index})") # pragma: no cover
        if same_exception:
            assert type(e) == type(exception), (e, exception)
            assert e.args == exception.args, (e.args, exception.args)
    else:
        if exception:
            fail(f"ndindex form did not raise but raw form does ({exception!r}): {index})") # pragma: no cover

    if not exception:
        assert_equal(a_raw, a_ndindex)


def iterslice(start_range=(-10, 10),
               stop_range=(-10, 10),
               step_range=(-10, 10),
               one_two_args=True
):
    # one_two_args is unnecessary if the args are being passed to slice(),
    # since slice() already canonicalizes missing arguments to None. We do it
    # for Slice to test that behavior.
    if one_two_args:
        for start in chain(range(*start_range), [None]):
            yield (start,)

        for start in chain(range(*start_range), [None]):
            for stop in chain(range(*stop_range), [None]):
                yield (start, stop)

    for start in chain(range(*start_range), [None]):
        for stop in chain(range(*stop_range), [None]):
            for step in chain(range(*step_range), [None]):
                yield (start, stop, step)


chunk_shapes = short_shapes

@composite
def chunk_sizes(draw, shapes=chunk_shapes):
    shape = draw(shapes)
    return draw(tuples(integers(1, 10), min_size=len(shape),
                       max_size=len(shape)).filter(lambda shape: prod(shape) < MAX_ARRAY_SIZE))



"""
Custom script to run the doctests

This runs the doctests but ignores trailing ``` in Markdown documents.

Running this separately from pytest also allows us to not include the doctests
in the coverage. It also allows us to force a separate namespace for each
docstring's doctest, which the pytest doctest integration does not allow.

TODO: Make these tests also run with pytest, but still keeping them out of the
coverage.

WARNING: This file registers these flags globally in the standard library
doctest module and monkeypatches the DoctestRunner object. It should only be
run as a standalone script. Do not attempt to import it.

"""

import numpy

import sys
import unittest
import glob
import os
from contextlib import contextmanager
from doctest import (DocTestRunner, DocFileSuite, DocTestSuite,
                     NORMALIZE_WHITESPACE, ELLIPSIS, register_optionflag)
import doctest

SKIPNP1 = register_optionflag("SKIPNP1")
NP1 = numpy.__version__.startswith('1')
if NP1:
    SKIP_THIS_VERSION = SKIPNP1
else:
    SKIP_THIS_VERSION = 0

@contextmanager
def patch_doctest():
    """
    Context manager to patch the doctester

    The doctester must be patched
    """
    orig_run = DocTestRunner.run
    orig_indent = doctest._indent

    def run(self, test, **kwargs):
        filtered_examples = []

        for example in test.examples:
            if SKIP_THIS_VERSION not in example.options:
                filtered_examples.append(example)

            # Remove ```
            example.want = example.want.replace('```\n', '')
            example.exc_msg = example.exc_msg and example.exc_msg.replace('```\n', '')

        test.examples = filtered_examples
        return orig_run(self, test, **kwargs)

    # Doctest indents the output, which is annoying for copy-paste, so disable
    # it.
    def _indent(s, **kwargs):
        return s

    try:
        DocTestRunner.run = run
        doctest._indent = _indent
        yield
    finally:
        DocTestRunner.run = orig_run
        doctest._indent = orig_indent

DOCS = os.path.realpath(os.path.join(__file__, os.path.pardir, os.path.pardir,
                                     os.pardir, 'docs'))
MARKDOWN = glob.glob(os.path.join(DOCS, '**', '*.md'), recursive=True)
RST = glob.glob(os.path.join(DOCS, '**', '*.rst'), recursive=True)
README = os.path.realpath(os.path.join(__file__, os.path.pardir, os.path.pardir,
                                     os.pardir, 'README.md'))
def load_tests(loader, tests, ignore):
    for mod in sys.modules:
        if mod.startswith('ndindex'):
            # globs={} makes the doctests not include module names
            tests.addTests(DocTestSuite(sys.modules[mod], globs={},
                                        optionflags=NORMALIZE_WHITESPACE))
    tests.addTests(DocFileSuite(*MARKDOWN, *RST, README,
                                optionflags=NORMALIZE_WHITESPACE | ELLIPSIS,
                                module_relative=False))
    return tests

def run_doctests():
    numpy.seterr(all='ignore')
    with patch_doctest():
        return unittest.main(module='ndindex.tests.doctest', exit=False).result

if __name__ == '__main__':
    # TODO: Allow specifying which doctests to run at the command line
    run_doctests()
